﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;             //【Debug】
using System.Drawing;                 //【Bitmap】【Color】
using System.Drawing.Imaging;         //【BitmapData】
using System.Runtime.InteropServices; //【Marshal.Copy】
using System.Threading;           //【Threadを使う為に必要】
using Microsoft.Kinect;           //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択
using Microsoft.DirectX;          //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒参照タブ
using Microsoft.DirectX.Direct3D; // DirectXを使う為
//【ディレクトリ】C:\Windows\Microsoft.NET\DirectX for Managed Code\1.0.2902.0\
//【選択dll】Microsoft.DirectX.dll, Microsoft.DirectX.Direct3D.dll, Microsoft.DirectX.Direct3DX.dll

//****************************************************************************
//【私設研究所Neo-Tech-Lab.com】『KINECT SENSORデモ・プログラム集』
//【プログラム】『3D-Player』【Cyber Space】
//【作者】上田智章
// All Rights reserved.  Copyright 2012  by ☆Tomoaki Ueda☆
//
//【Microsoft Visual Studio 2010 Express】 Visual C# 2010 Express
//【Microsoft KINECT for Windows SDK Version1】
//【使用許諾事項】
//(1)本プログラムの使用、ソースコードの一部または全部の利用によって生じた
//　いかなる不利益に対しても、著者およびＣＱ出版株式会社が補償することは
//　ありません。自己責任の下での使用及び利用を行わなければなりません。
//****************************************************************************
//【内容】Kinect for Windowsは、Frameworks Ver.4.0でのみ動作可能です。
//　ところが、Managed DirectXはFrameworks Ver.4.0上ではデフォルトで動作しない
//  ので、App.configファイルで設定をしてやる必要があります。
//  このプログラムでは、背景画をSurfaceで書き換えてから２つの正方形を描画する
//  サンプル・プログラムです。
//****************************************************************************

namespace D3D_RGBCamera
{
    public partial class Form1 : Form
    {
        //【KINECT関連】-------pictureBox1の大きさ------------------------------------------
        private const int KS_ScreenWidth = 640;  //【KINECTのImageFrame横方向画素数】 
        private const int KS_ScreenHeight = 480; //【KINECTのImageFrame縦方向画素数】
        private KinectSensor KS_KinectSensor;    //【KINECT】
        private byte[] KS_ColorImageDataXRGB;    //【KINECTのColorImageData】32bit/画素,X8R8G8B8表示形式
        private DepthImagePixel[] KS_DepthData;  //【KINECTのDepthImageData】16bit/画素,Depth 13bits, PlayerID 3bts表示形式
        private ColorImagePoint[] KS_ColorPoint; //
        private Skeleton[] SkeletonData;         //【Kinectの骨格データ】
        private Bitmap KS_bmp1;                  //【ColorImage表示用Bitmapオブジェクト】
        private BitmapData KS_data1;             //【ColorImage用BitmapData】
        private Rectangle KS_rect1;              //【ColorImage表示枠】
        private bool KS_flagFlip = true;         //【左右反転フラグ】
        private ColorBar KS_ColorBar = new ColorBar(); //【デプス値をカラーに変換する部品】

        public struct Vector
        {
            public bool T;
            public float X;
            public float Y;
            public float Z;
        }

        public Vector[] JointPosition = new Vector[20]; //【KINECTセンサのSkeletonPointはPlayer当り20点】
        public JointType[] JointList = //列挙型が面倒なので通常のVector3配列に変換する為のデータ
        {  JointType.Head, JointType.ShoulderCenter, JointType.Spine, JointType.HipCenter,
           JointType.ShoulderLeft, JointType.ElbowLeft, JointType.WristLeft, JointType.HandLeft,
           JointType.ShoulderRight,JointType.ElbowRight,JointType.WristRight,JointType.HandRight,
           JointType.HipLeft,JointType.KneeLeft,JointType.AnkleLeft,JointType.FootLeft,
           JointType.HipRight,JointType.KneeRight,JointType.AnkleRight,JointType.FootRight
        };

        //【DirectX関連】-------------------------------------------------------------------
        private Device D3Ddevice = null; //【Direct3Dデバイス】
        private Vector3 D3DcameraPosition = new Vector3(-10.0f, 10.0f, -150.0f); //【カメラの位置】
        private Vector3 D3DcameraEyeUp = new Vector3(0.0f, 0.0f, 1.0f);        //【Y軸方向が上】
        private Vector3 D3DcameraTarget = new Vector3(0.0f, 1.0f, 0.0f);       //【注視点座標】
        private Vector3 D3Dlight = new Vector3(-0.57735f, -0.57735f, -0.57735f); //【光源方向】
        private VertexBuffer D3DvtxBuffer = null; //【頂点バッファ】球体の頂点情報(座標と色)を格納するバッファ
        private IndexBuffer D3DidxBuffer = null;  //【インデックスバッファ】球体を構成する三角形の頂点番号を格納するバッファ
        private Surface D3Dsurface;
        private Surface D3Dbackbuffer;
        private float AngleX = 0f; //【座標回転角度】
        private float AngleY = 0f; //【座標回転角度】
        private float Omega = (float)Math.PI / 180.0f;
        private Microsoft.DirectX.Matrix m1 = new Microsoft.DirectX.Matrix();  //【行列】
        private Microsoft.DirectX.Matrix m2 = new Microsoft.DirectX.Matrix();  //【行列】
        private Microsoft.DirectX.Matrix m3 = new Microsoft.DirectX.Matrix();  //【行列】

        //【NTL拡張現実センサ関連】---------------------------------------------------------
        private Thread processThread;            //【MainLoop用Thread】
        private bool processRun;                 //Thread実行フラグ

        public Form1()
        {
            InitializeComponent();

            //***********************************************************
            //【Kinect検出】【共通】
            //***********************************************************
            //【利用可能なKINECTセンサを取得する】
            KS_KinectSensor = (from sensorToCheck
                               in KinectSensor.KinectSensors
                               where sensorToCheck.Status == KinectStatus.Connected
                               select sensorToCheck).FirstOrDefault();
            //【検出結果の表示】
            pictureBox2.SizeMode = PictureBoxSizeMode.StretchImage;
            if (KS_KinectSensor == null) //【Kinect接続が確認できなかった場合】
            {
                pictureBox2.Image = new Bitmap("../../image/KinectOff.png");
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            else //【Kinect検出に成功】
            {
                pictureBox2.Image = new Bitmap("../../image/KinectOn.png");
            }
            KS_KinectSensor.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30);
            KS_KinectSensor.DepthStream.Enable(DepthImageFormat.Resolution640x480Fps30);
            KS_KinectSensor.SkeletonStream.Enable();
            KS_KinectSensor.Start();
            KS_KinectSensor.ColorFrameReady += new EventHandler<ColorImageFrameReadyEventArgs>(KS_KinectSensor_ColorFrameReady);
            KS_KinectSensor.DepthFrameReady += new EventHandler<DepthImageFrameReadyEventArgs>(KS_KinectSensor_DepthFrameReady);
            KS_KinectSensor.AllFramesReady += new EventHandler<AllFramesReadyEventArgs>(KS_KinectSensor_AllFramesReady);
            KS_ColorImageDataXRGB = new byte[4 * KS_ScreenWidth * KS_ScreenHeight];
            KS_DepthData = new DepthImagePixel[KS_ScreenWidth * KS_ScreenHeight];
            KS_ColorPoint = new ColorImagePoint[KS_ScreenWidth * KS_ScreenHeight];
            KS_bmp1 = new Bitmap(KS_ScreenWidth, KS_ScreenHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【DepthImage表示用】
            KS_rect1 = new Rectangle(0, 0, KS_ScreenWidth, KS_ScreenHeight); //【DepthImage用枠】
            KS_ColorBar.CreateColorPalletForDistance();
            pictureBox3.Image = KS_ColorBar.CreateColorScale(512, 10, false);
            pictureBox3.SizeMode = PictureBoxSizeMode.StretchImage;
            KS_KinectSensor.DepthStream.Range = DepthRange.Near;
            KS_KinectSensor.CoordinateMapper.MapDepthFrameToColorFrame(DepthImageFormat.Resolution640x480Fps30, KS_DepthData, ColorImageFormat.RgbResolution640x480Fps30, KS_ColorPoint);
            SkeletonData = new Skeleton[6];

            pictureBox1.ClientSize = new Size(KS_ScreenWidth, KS_ScreenHeight);
            D3Ddevice = PrepareD3dDevice(this.pictureBox1); //【DirectX初期化】device
            processRun = true;                              //【Thread実行フラグ】

            button1.Click+=new EventHandler(button1_Click);

            CustomVertex.PositionColored[] NTL_vtxData = new CustomVertex.PositionColored[8]; //8点の情報を格納するためのメモリを確保
            const float CUBE_SIZE = 20.0f; //1辺40[mm]の頂点を設定
            NTL_vtxData[0] = new CustomVertex.PositionColored(-CUBE_SIZE, CUBE_SIZE, CUBE_SIZE, Color.Yellow.ToArgb());
            NTL_vtxData[1] = new CustomVertex.PositionColored(CUBE_SIZE, CUBE_SIZE, CUBE_SIZE, Color.Gray.ToArgb());
            NTL_vtxData[2] = new CustomVertex.PositionColored(-CUBE_SIZE, CUBE_SIZE, -CUBE_SIZE, Color.Purple.ToArgb());
            NTL_vtxData[3] = new CustomVertex.PositionColored(CUBE_SIZE, CUBE_SIZE, -CUBE_SIZE, Color.Red.ToArgb());
            NTL_vtxData[4] = new CustomVertex.PositionColored(-CUBE_SIZE, -CUBE_SIZE, CUBE_SIZE, Color.SkyBlue.ToArgb());
            NTL_vtxData[5] = new CustomVertex.PositionColored(CUBE_SIZE, -CUBE_SIZE, CUBE_SIZE, Color.Orange.ToArgb());
            NTL_vtxData[6] = new CustomVertex.PositionColored(-CUBE_SIZE, -CUBE_SIZE, -CUBE_SIZE, Color.Green.ToArgb());
            NTL_vtxData[7] = new CustomVertex.PositionColored(CUBE_SIZE, -CUBE_SIZE, -CUBE_SIZE, Color.Blue.ToArgb());
            Int16[] NTL_idxData = new Int16[] { 2, 0, 1, 1, 3, 2, 4, 0, 2, 2, 6, 4, 5, 1, 0, 0, 4, 5, 7, 3, 1, 1, 5, 7, 6, 2, 3, 3, 7, 6, 4, 6, 7, 7, 5, 4 };

            //立方体（頂点数8）の準備
            D3DvtxBuffer = new VertexBuffer(typeof(CustomVertex.PositionColored), 8, D3Ddevice, Usage.None, CustomVertex.PositionColored.Format, Pool.Managed);
            using (GraphicsStream data = D3DvtxBuffer.Lock(0, 0, LockFlags.None)) //頂点バッファをロックする
            {
                data.Write(NTL_vtxData); // 頂点データを頂点バッファにコピーします
                D3DvtxBuffer.Unlock();// 頂点バッファのロックを解除します
            }
            D3DidxBuffer = new IndexBuffer(D3Ddevice, 12 * 3 * 2, Usage.WriteOnly, Pool.Managed, true);
            using (GraphicsStream data = D3DidxBuffer.Lock(0, 0, LockFlags.None))
            {
                data.Write(NTL_idxData); // インデックスデータをインデックスバッファにコピーします
                D3DidxBuffer.Unlock(); // インデックスバッファのロックを解除します
            }

            D3Dbackbuffer = D3Ddevice.GetBackBuffer(0, 0, BackBufferType.Mono);
            D3Dsurface = D3Ddevice.CreateOffscreenPlainSurface(KS_ScreenWidth, KS_ScreenHeight, Format.X8R8G8B8, Pool.Default); //画像を格納するサーフェイスを作成
            SurfaceLoader.FromFile(D3Dsurface, "../../image/Sky.bmp", Filter.None, 0x000000); //【ファイルから元画像を入れる】

            D3Ddevice.RenderState.ZBufferEnable = true;                 //【陰面処理をＺバッファ方式で行う】
            D3Ddevice.RenderState.ZBufferFunction = Compare.LessEqual;  //【手前がＺ値の小さい側】.LessEqual;    
            D3Ddevice.RenderState.Lighting = false;                     //【光源計算処理を行う】true;
            D3Ddevice.RenderState.CullMode = Cull.CounterClockwise;

            //【ビュー変換の設定】(左手座標系ビュー行列で設定する) (0,0,0)から、Z+方向を向いて、上方向がY軸
            D3Ddevice.Transform.View = Microsoft.DirectX.Matrix.LookAtLH(D3DcameraPosition, D3DcameraEyeUp, D3DcameraTarget);
            D3Ddevice.Transform.Projection = Microsoft.DirectX.Matrix.PerspectiveFovLH((float)Math.PI / 4.0f, 1.0f, 0.001f, 1000.0f);
            //【パラメータ】視野角45°, 画面のアスペクト比1.0f, 最も近い距離1.0f, 最も遠い距離100.0f
            //【ビューポート設定】
            Viewport vp = new Viewport();
            vp.X = 0; vp.Y = 0; vp.Height = KS_ScreenHeight; vp.Width = KS_ScreenWidth; vp.MaxZ = 1.0f;
            D3Ddevice.Viewport = vp;

            processThread = new Thread(MainProcessThread);  //【Main処理ループのためのThreadを定義】
            processThread.Start();                          //【Thread開始】
        }


        //***************************************************************
        //【拡張現実３次元表示スレッド】
        //***************************************************************
        void MainProcessThread()
        {
            while(processRun)
            {
                KS_ConvertDepthToColor();
                KS_data1 = KS_bmp1.LockBits(KS_rect1, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【Lockする】
                Marshal.Copy(KS_ColorImageDataXRGB, 0, KS_data1.Scan0, KS_ColorImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列depthDataをコピーする】
                KS_bmp1.UnlockBits(KS_data1); //【Lock解除】
                if (KS_flagFlip) { KS_bmp1.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                Surface src = Surface.FromBitmap(D3Ddevice, KS_bmp1, Pool.Default);
                SurfaceLoader.FromSurface(D3Dsurface, src, Filter.None, 0x000000); //【ファイルから元画像を入れる】

                D3Ddevice.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.DarkBlue, 1.0f, 0);
                D3Ddevice.BeginScene(); //【シーン開始】
                try
                {
                    D3Ddevice.StretchRectangle(D3Dsurface, new Rectangle(0, 0, 640, 480), D3Dbackbuffer, new Rectangle(0, 0, 640, 480), TextureFilter.None);
                }
                catch (Exception em) { Debug.Write(em.ToString()); }

                //【ワールド、ビュー、射影行列を設定】
                m1 = Microsoft.DirectX.Matrix.RotationY(AngleX * Omega);  //【y軸回りに15°回転】
                m1 *= Microsoft.DirectX.Matrix.RotationX(AngleY * Omega); //【x軸回りに25°回転】
                AngleX += 1.5f; if (AngleX > 360.0f) { AngleX -= 360.0f; }
                AngleY += 0.9f; if (AngleX > 360.0f) { AngleY -= 360.0f; }
                D3Ddevice.Transform.World = Microsoft.DirectX.Matrix.Identity;

                D3Ddevice.SetStreamSource(0, this.D3DvtxBuffer, 0);
                D3Ddevice.VertexFormat = CustomVertex.PositionColored.Format;
                D3Ddevice.Indices = D3DidxBuffer; // インデックスバッファをセット
                m2 = m1 * Microsoft.DirectX.Matrix.Translation(-30.0f, 95.0f, 230.0f);
                D3Ddevice.SetTransform(TransformType.World, m2); // 計算したマトリックスで座標変換
                D3Ddevice.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, 8, 0, 12);
                m3 = m1 * Microsoft.DirectX.Matrix.Translation(-40.0f, -30.0f, 20.0f);
                D3Ddevice.SetTransform(TransformType.World, m3); // 計算したマトリックスで座標変換
                D3Ddevice.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, 8, 0, 12);
                D3Ddevice.EndScene();
                D3Ddevice.Present();

                Application.DoEvents();
                this.Invalidate();
            }
            D3Ddevice.Dispose();
            D3DidxBuffer.Dispose();
            D3DidxBuffer.Dispose();
            D3Dbackbuffer.Dispose();
            D3Dsurface.Dispose();
            this.Close();
        }

        //***************************************************************
        //【拡張現実３次元表示スレッド終了フラグ設定】
        //***************************************************************
        private void button1_Click(object sender, EventArgs e)
        {
            processRun = false;
        }

        //***************************************************************
        //【ColorFrame取得イベント】
        //***************************************************************
        void KS_KinectSensor_ColorFrameReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (ColorImageFrame f = e.OpenColorImageFrame())
            {
                if(f != null) { f.CopyPixelDataTo(KS_ColorImageDataXRGB); }
            }
        }

        //***************************************************************
        //【DepthFrame取得イベント】
        //***************************************************************
        void KS_KinectSensor_DepthFrameReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            using (DepthImageFrame f = e.OpenDepthImageFrame())
            {
                if (f != null) { f.CopyDepthImagePixelDataTo(KS_DepthData); }
            }
        }

        //***************************************************************
        //【SkeletonFrame取得イベント】
        //***************************************************************
        void KS_KinectSensor_AllFramesReady(object sender, AllFramesReadyEventArgs e)
        {
            using (SkeletonFrame f = e.OpenSkeletonFrame())
            {
                if (f != null) {
                    f.CopySkeletonDataTo(SkeletonData);
                    JointPosition[0].T = false;
                    JointPosition[1].T = false;
                    for (int i = 0; i < SkeletonData.Length; i++)
                    {
                        if ((SkeletonData[i].TrackingState == SkeletonTrackingState.Tracked)
                         && (SkeletonData[i].Joints[JointType.HandLeft].TrackingState==JointTrackingState.Tracked))
                        {
                            JointPosition[0].X = SkeletonData[i].Joints[JointType.HandLeft].Position.X;
                            JointPosition[0].Y = SkeletonData[i].Joints[JointType.HandLeft].Position.Y;
                            JointPosition[0].Z = SkeletonData[i].Joints[JointType.HandLeft].Position.Z;
                            JointPosition[0].T = true;
                        }
                        if ((SkeletonData[i].TrackingState == SkeletonTrackingState.Tracked)
                         && (SkeletonData[i].Joints[JointType.HandRight].TrackingState == JointTrackingState.Tracked))
                        {
                            JointPosition[1].X = SkeletonData[i].Joints[JointType.HandLeft].Position.X;
                            JointPosition[1].Y = SkeletonData[i].Joints[JointType.HandLeft].Position.Y;
                            JointPosition[1].Z = SkeletonData[i].Joints[JointType.HandLeft].Position.Z;
                            JointPosition[1].T = true;
                        }
                        if (JointPosition[0].T && JointPosition[1].T) { break; }
                    }
                }
            }
        }




        //***************************************************************
        //【ConvertDepthToColor】16ビットのDepth値をカラーバーに対応する色に変更
        //***************************************************************
        private void KS_ConvertDepthToColor()
        {
            int i, j, d, L;
            L = 4 * 640 * 480;

            for (i = 0; i < KS_DepthData.Length; i++)
            {
                if (KS_DepthData[i].PlayerIndex > 0)
                {
                    d = KS_DepthData[i].Depth;
                    if (d > 4000) { d = 4000; }
                    j = 4 * (KS_ColorPoint[i].X + 640 * KS_ColorPoint[i].Y);
                    if((j>=0) && (j<L))
                    KS_ColorImageDataXRGB[j] = (byte)((KS_ColorImageDataXRGB[j] + KS_ColorBar.B[d]) * 0.5f); //【青】B
                    KS_ColorImageDataXRGB[j + 1] = (byte)((KS_ColorImageDataXRGB[j + 1] + KS_ColorBar.G[d])*0.5f); //【緑】G
                    KS_ColorImageDataXRGB[j + 2] = (byte)((KS_ColorImageDataXRGB[j + 2] + KS_ColorBar.R[d])*0.5f); //【赤】R
                    KS_ColorImageDataXRGB[j + 3] = (byte)((KS_ColorImageDataXRGB[j + 3] + KS_ColorBar.A[d])*0.5f); //【α】A
                }
            }
        }

        //***************************************************************
        //【Direct3Dデバイスを準備する関数】
        //***************************************************************
        private Device PrepareD3dDevice(Control i_window)
        {
            PresentParameters pp = new PresentParameters(); //【プレゼンテーション・パラメータを作成】
            pp.Windowed = true;                             //【ウィンドウモード】pictureBox1を表示域に設定
            pp.SwapEffect = SwapEffect.Discard;             //【スワップエフェクトを設定】
            pp.BackBufferFormat = Format.X8R8G8B8;          //【描画用グラフィックメモリ】RGB各8bit, 1670万色
            pp.BackBufferCount = 1;                         //【バッファの数】1
            pp.BackBufferWidth = KS_ScreenWidth;            //
            pp.BackBufferHeight = KS_ScreenHeight;          //
            pp.EnableAutoDepthStencil = true;               //【Ｚバッファを使用する】
            pp.AutoDepthStencilFormat = DepthFormat.D16;    //【Ｚバッファ16bit】
            CreateFlags fl_base = CreateFlags.FpuPreserve;  //【初期化に必要なフラグ】

            try { return new Device(0, DeviceType.Hardware, i_window.Handle, fl_base | CreateFlags.HardwareVertexProcessing, pp); }
            catch (Exception ex1)
            {
                Debug.WriteLine(ex1.ToString());
                try { return new Device(0, DeviceType.Hardware, i_window.Handle, fl_base | CreateFlags.SoftwareVertexProcessing, pp); }
                catch (Exception ex2)
                {
                    Debug.WriteLine(ex2.ToString());
                    try { return new Device(0, DeviceType.Reference, i_window.Handle, fl_base | CreateFlags.SoftwareVertexProcessing, pp); }
                    catch (Exception ex3) { throw ex3; }
                }
            }
        }

        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }
        }


    }
}